"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const graphql_1 = require("graphql");
const ts = require("typescript");
const date_1 = require("./date");
const extractor_1 = require("./extractor");
function createSchema(fileName, options = {}) {
    const customScalarsMap = new Map([date_1.DateType, ...(options.customScalars || [])].map(it => [it.name, it]));
    const program = ts.createProgram({ options: { strict: true }, rootNames: [fileName] });
    const sourceFile = program.getSourceFile(fileName);
    const types = extractor_1.extractTypes(program)(sourceFile);
    const map = new Map();
    let anonTypeIdx = 0;
    const schema = createSchemaFromTypes();
    // console.log(printSchema(schema));
    return schema;
    function createSchemaFromTypes() {
        let query;
        let mutation;
        for (let i = 0; i < types.length; i++) {
            const gqlType = createGQL(types[i]);
            if (gqlType instanceof graphql_1.GraphQLObjectType) {
                if (gqlType.name.match(/Query$/)) {
                    query = gqlType;
                }
                if (gqlType.name.match(/Mutation$/)) {
                    mutation = gqlType;
                }
            }
        }
        return new graphql_1.GraphQLSchema({
            query: query,
            mutation: mutation,
        });
    }
    function add(type, gqltype) {
        map.set(type, gqltype);
        return gqltype;
    }
    function createGQL(type) {
        const gqlType = map.get(type);
        if (gqlType)
            return gqlType;
        switch (type.kind) {
            case 'interface':
                return createGQLType(type);
            case 'union':
                return add(type, createGQLUnion(type));
            case 'native':
                if (type.type === 'array') {
                    return new graphql_1.GraphQLList(add(type, nullable(false, createGQL(type.element))));
                }
                throw new Error('Unexpected type: ' + type.type);
            case 'primitive':
                return add(type, createGQLPrimitive(type));
        }
        throw new Error('Unexpected type: ' + type.kind);
    }
    function nullable(nullable, type) {
        return nullable ? type : new graphql_1.GraphQLNonNull(type);
    }
    function createGQLType(type) {
        const isInput = type.name.match(/Input$/);
        const Class = isInput ? graphql_1.GraphQLInputObjectType : graphql_1.GraphQLObjectType;
        const fields = {};
        let name = type.name;
        if (type.name === 'anonymous') {
            const __typename = type.members.find(member => member.name === '__typename');
            name =
                (__typename && __typename.type.literal) || 'Anonymous' + ++anonTypeIdx;
        }
        const gqlType = new Class({
            name: name,
            description: type.doc,
            fields: fields,
        });
        add(type, gqlType);
        type.members.reduce((obj, member) => {
            const memberType = {
                type: nullable(member.optional, createGQL(member.type)),
                args: member.args
                    ? member.args.reduce((acc, arg) => {
                        acc[arg.name] = {
                            description: arg.doc,
                            defaultValue: arg.defaultValue,
                            type: nullable(arg.optional, createGQL(arg.type)),
                        };
                        return acc;
                    }, {})
                    : undefined,
                // todo:
                deprecationReason: undefined,
                description: member.doc,
            };
            if (member.name !== '__typename') {
                obj[member.name] = memberType;
            }
            return obj;
        }, fields);
        return gqlType;
    }
    function createGQLUnion(type) {
        return new graphql_1.GraphQLUnionType({
            name: type.name,
            description: type.doc,
            types: type.members.map(member => createGQL(member)),
        });
    }
    function createGQLPrimitive(type) {
        switch (type.type) {
            case 'Float':
            case 'number':
                return graphql_1.GraphQLFloat;
            case 'Int':
                return graphql_1.GraphQLInt;
            case 'ID':
                return graphql_1.GraphQLID;
            case 'string':
                return graphql_1.GraphQLString;
            case 'boolean':
                return graphql_1.GraphQLBoolean;
        }
        const customType = customScalarsMap.get(type.type);
        if (customType)
            return customType;
        throw new Error('Unexpected type: ' + type.type);
    }
}
exports.createSchema = createSchema;
