"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
function extractTypes(program) {
    const checker = program.getTypeChecker();
    const types = new Map();
    let id = 0;
    return (sourceFile) => {
        function getType(tsType, rawType) {
            let type = types.get(tsType);
            if (type)
                return type;
            const symbol = tsType.symbol;
            const isString = tsType.flags & ts.TypeFlags.String;
            const isNumber = tsType.flags & ts.TypeFlags.Number;
            const isBoolean = tsType.flags & ts.TypeFlags.Boolean;
            if (isString || isNumber || isBoolean) {
                const type = {
                    kind: 'primitive',
                    type: rawType || checker.typeToString(tsType),
                    literal: undefined,
                };
                return type;
            }
            if (tsType.isLiteral()) {
                const type = {
                    kind: 'primitive',
                    type: rawType || isString ? 'string' : isNumber ? 'number' : isBoolean ? 'boolean' : 'other',
                    literal: tsType.value,
                };
                return type;
            }
            tsType = tsType.getNonNullableType();
            if (checker.isArrayLikeType(tsType)) {
                const type = {
                    kind: 'native',
                    type: 'array',
                    element: getType(tsType.typeArguments[0]),
                };
                return type;
            }
            if (symbol && symbol.name === 'Date') {
                const type = {
                    kind: 'primitive',
                    type: 'Date',
                    literal: undefined,
                };
                return type;
            }
            if (tsType.flags & ts.TypeFlags.Object && tsType.symbol.flags & ts.SymbolFlags.TypeLiteral) {
                const type = {
                    id: id++,
                    kind: 'interface',
                    name: 'anonymous',
                    doc: getDoc(symbol),
                    members: checker.getPropertiesOfType(tsType).map(createProp),
                };
                types.set(tsType, type);
                return type;
            }
            type = { id: id++ };
            types.set(tsType, type);
            return type;
        }
        function getDoc(symbol) {
            if (!symbol)
                return;
            const doc = symbol.getDocumentationComment(checker);
            if (doc.length > 0)
                return doc[0].text;
        }
        function createProp(symbol) {
            const tsType = getTypeFromSymbol(symbol);
            const signature = checker.getSignaturesOfType(tsType, ts.SignatureKind.Call)[0];
            let paramsType;
            if (signature) {
                const paramsArgSymbol = signature.parameters[0];
                if (paramsArgSymbol) {
                    paramsType = getTypeFromSymbol(paramsArgSymbol);
                }
            }
            const retType = signature ? signature.getReturnType() : tsType;
            const declNode = symbol.declarations[0].type;
            let rawType = declNode.getText();
            let defaultValue;
            if (ts.isTypeReferenceNode(declNode) &&
                ts.isIdentifier(declNode.typeName) &&
                declNode.typeName.text === 'Default') {
                const defaultValueNode = declNode.typeArguments[1];
                if (ts.isLiteralTypeNode(defaultValueNode)) {
                    if (ts.isStringLiteral(defaultValueNode.literal) || ts.isNumericLiteral(defaultValueNode.literal)) {
                        defaultValue = defaultValueNode.literal.text;
                    }
                    if (defaultValueNode.literal.kind === ts.SyntaxKind.TrueKeyword) {
                        defaultValue = true;
                    }
                    if (defaultValueNode.literal.kind === ts.SyntaxKind.FalseKeyword) {
                        defaultValue = false;
                    }
                }
                rawType = declNode.typeArguments[0].getText();
            }
            return {
                name: symbol.name,
                doc: getDoc(symbol),
                type: getType(retType, rawType),
                defaultValue: defaultValue,
                args: paramsType ? paramsType.getProperties().map(createProp) : undefined,
                optional: (symbol.flags & ts.SymbolFlags.Optional) > 0 || retType.getNonNullableType() !== retType,
            };
        }
        function getTypeFromSymbol(symbol) {
            return checker.getTypeOfSymbolAtLocation(symbol, symbol.declarations[0]);
        }
        function visitor(node) {
            if (ts.isInterfaceDeclaration(node)) {
                const tsType = checker.getTypeAtLocation(node);
                // const tsType = checker.getTypeOfSymbolAtLocation(symbol, symbol.declarations[0]);
                const type = getType(tsType);
                type.doc = getDoc(tsType.symbol);
                type.kind = 'interface';
                type.name = node.name.text;
                type.members = [];
                const symbols = checker.getPropertiesOfType(tsType);
                for (let i = 0; i < symbols.length; i++) {
                    const symbol = symbols[i];
                    type.members.push(createProp(symbol));
                }
            }
            if (ts.isTypeAliasDeclaration(node)) {
                if (ts.isUnionTypeNode(node.type)) {
                    const tsType = checker.getTypeAtLocation(node.type);
                    const type = getType(tsType);
                    type.doc = getDoc(tsType.aliasSymbol);
                    type.kind = 'union';
                    type.name = node.name.text;
                    type.members = node.type.types.map(typeNode => getType(checker.getTypeFromTypeNode(typeNode)));
                }
            }
        }
        ts.forEachChild(sourceFile, visitor);
        return [...types.values()];
    };
}
exports.extractTypes = extractTypes;
